"""Package providing rclone file repositories.

Author: Bernd Kalbfuss
License: GNU General Public License v3 (GPLv3)
"""

from . file import RepositoryFile
from . repository import Repository, FileIterator
